"""Classes for constructing maps of scientific literature.

The `sciterra.mapping.atlas` submodule contains the basic data structure, the Atlas.

The `sciterra.mapping.cartography` submodule contains functionality for manipulating an Atlas.

"""
